from math import ceil
from itertools import zip_longest
from os.path import commonprefix
from itertools import cycle
from array import array


__all__ = (
    "DataSource",
    "Preprocessor",
    "LearnblockIdentifier"
)


class DataSource:
    _SIGNED_CHAR = 'b'

    def __init__(self, source, learnblock_identifier, settings):
        self.settings = settings
        self.n_source = 0
        self.features_before_mapping = None
        self.features_after_mapping = None
        self.source = source
        # self.__learnblock_identifier = learnblock_identifier
        self.identifier = learnblock_identifier
        self.source_halde_flags = None

    def __getstate__(self):
        state = dict(source=self.source,
                     features_before_mapping=self.features_before_mapping,
                     features_after_mapping=self.features_after_mapping,
                     settings=self.settings,
                     n_source=self.n_source)
        return state

    def __setstate__(self, state):
        self.settings = state.get("settings")
        self.n_source = state.get("n_source")
        self.source = state.get("source")
        self.features_before_mapping = state.get("features_before_mapping")
        self.features_after_mapping = state.get("features_after_mapping")

    @property
    def blocks(self):
        if self.settings.block_size > len(self):
            raise ValueError("Block size cannot be larger then the size"
                             "the data source.")
        for block in self:
            yield block

    @property
    def learnblocks(self):
        if self.settings.block_size > len(self):
            raise ValueError("Block size cannot be larger then the size"
                             "the data source.")

        # for block in self:
        #     learnblock = self.__learnblock_identifier.identify(block)
        #     if learnblock:
        #         learnblock.origin = ("source", self.n_source)
        #         self._flip_source_halde_flags(learnblock.indices())
        #         yield learnblock
        for block in self:
            yield block

    def __next__(self):
        return next(self.__next_helper)

    def _next_helper(self):
        block_indexes = []
        counter = 0
        old_index = 0
        halde_runs = -1

        for i in cycle(range(0, len(self))):
            if all(self.source_halde_flags):
                # all data was selected
                return

            if halde_runs >= self.settings.stack_iterations:
                # manually stop generator
                return

            if counter == self.settings.block_size:
                old_index = i
                counter = 0
                yield self.source.new_block_from(block_indexes, index=True)
                block_indexes.clear()

            elif old_index > i:
                halde_runs += 1
                old_index = i
                counter = 0
                yield self.source.new_block_from(block_indexes, index=True)
                block_indexes.clear()

            if self.source_halde_flags[i] == 0:
                block_indexes.append(i)
                counter += 1

    def __iter__(self):
        self.__next_helper = iter(self._next_helper())
        self.source_halde_flags = array(
            self._SIGNED_CHAR, [0 for _ in range(len(self))])
        return self

    def _flip_source_halde_flags(self, learnblock_indexes):
        for i in learnblock_indexes:
            self.source_halde_flags[i] = 1

    def __len__(self):
        return len(self.source)

    def get_block(self, indices=None, columns=None):
        return self.source.new_block_from(indices, index=True)

    def time_sigma_relatives(self, block):
        return next(iter(self.identifier.identify_relatives(
            block, "T", "Sigma")))

    def estimate_density(self, data):
        kernel_density_estimator = self.identifier.density_estimator.train(data)
        return kernel_density_estimator.density(data)

    def remove_time_dense_relatives(self, density):
        return self.identifier._remove_time_dense_relatives(density)

    def cluster(self, block, density):
        return self.identifier._cluster_sigma_zeta_relatives(block, density)

    def new_learnblock(self, values, columns, index, origin):
        return self.source.new_learnblock(values, columns, index, origin)

    def get_time_values(self, indices):
        return self.source.get_block_via_index(indices, columns="T")\
            .as_numpy_array()

    def remap_feature_names(self):
        self.features_before_mapping = self.source.columns(effective=True)
        count_columns = self.source.n_columns(effective=True)
        new_column_names = []
        for n_feature in range(count_columns):
            name = ".".join([str(self.n_source), "0", str(n_feature)])
            new_column_names.append(name)

        self.source.rename_columns(new_column_names)
        self.features_after_mapping = self.source.columns(effective=False)

    def reindex(self, start, end):
        self.source.reindex(range(start, end))


class Preprocessor:
    TIME_COLUMN = "T"
    TARGET_COLUMN = "Z"
    SIGMA_COLUMN = "Sigma"

    def __init__(self, settings):
        self.settings = settings

    def clean(self, table):
        self._set_proper_column_names(table)
        self._drop_irrelevant_columns(table)

        if self.settings.set_targets:
            self._overwrite_target_column(table)

        if self.settings.sort_time_stamp:
            self._sort_according_time_stamp(table)

        if self.settings.cut_time_stamp:
            self._remove_common_time_stamp_prefix(table)

        return table

    def _set_proper_column_names(self, table):
        names = []
        for num in range(table.n_columns(effective=True)):
            names.append(".".join(["0.0", str(num)]))
        table.rename_columns(names)

    def _drop_irrelevant_columns(self, table):
        features_to_be_removed = []
        for index in range(table.n_columns()):
            if index not in self.settings.set_features:
                features_to_be_removed.append(index)
        table.drop_columns(features_to_be_removed, by_index=True)

        # table.drop_columns(self.settings.set_features, )
        # features_to_be_removed = [table.get_column_name_by_index(i)
        #                           for i in self.settings.set_features]
        # for column in features_to_be_removed:
        #     table.drop_column_by_name(column)

    def _overwrite_target_column(self, table):
        table.set_column(self.TARGET_COLUMN, self.settings.set_targets)

    def _sort_according_time_stamp(self, table):
        table.sort_after(self.TIME_COLUMN)

    def _remove_common_time_stamp_prefix(self, table):
        # TODO (dmt): Check if timestamp column is of type string!
        time_column = table.get_column_values_as_list(self.TIME_COLUMN)
        common_prefix = commonprefix(time_column)
        cleaned_time_column = [s.lstrip(common_prefix) for s in time_column]
        table.set_column_values(cleaned_time_column)


class LearnblockIdentifier:
    def __init__(self, settings, density_estimator):
        self.settings = settings
        self.density_estimator = density_estimator

    @classmethod
    def _column_pairs(cls):
        yield ("Sigma", "Z")
        yield ("T", "Sigma")
        yield ("T", "Z")

    def identify(self, block):
        biggest_learn_block = None
        biggest_block_size = 0

        for pair in LearnblockIdentifier._column_pairs():
            for possible_learnblock in self._identify_relatives(block, *pair):
                if self._is_learn_block(possible_learnblock.n_rows()):
                    if possible_learnblock.n_rows() > biggest_block_size:
                        biggest_learn_block = possible_learnblock
                        biggest_learn_block.relationship = pair

        return biggest_learn_block

    def _is_learn_block(self, block_length):
        return block_length >= self.settings.learn_block_minimum

    def _identify_relatives(self, block, *args):
        already_seen = set()
        for value_pair in block.get_duplicated_pairs(args[0], args[1]):
            if value_pair not in already_seen:
                already_seen.add(value_pair)
                kw = {args[0]: value_pair[0], args[1]: value_pair[1]}
                if args[0] == "Sigma" and args[1] == "Z":
                    try:
                        for block in self._get_sigma_zeta_relatives(block,
                                                                    **kw):
                            yield block
                    except ValueError as e:
                        # TODO (dmt): Provide mechanism for logging errors!
                        print(kw)
                        continue

                else:
                    yield block.construct_from_two_pairs(**kw)

    def _get_sigma_zeta_relatives(self, block, **kw):
        relatives = block.construct_from_two_pairs(**kw)
        time_column = relatives.get_column("T")
        density = self.density_estimator.train(time_column).density(time_column)
        changes = self._remove_time_dense_relatives(density)
        clusters = self._cluster_sigma_zeta_relatives(relatives, changes)
        for time_values in clusters:
            yield relatives.new_block_from(time_values, time_column=True)
            # yield relatives.new_block_from(time_values)

    def _remove_time_dense_relatives(self, density):
        changes = list()
        max_dens = max(density)
        for index, dens in enumerate(density):
            if dens > max_dens*self.settings.sigma_zeta_cutoff:
                changes.append(True)
            else: changes.append(False)
        return changes

    def _cluster_sigma_zeta_relatives(self, cutted_block, changes):
        # TOOD (dmt): Don't rely on data series from pandas, 'cause ckmeans
        # needs primitives data types.
        count_changes = 0
        for i, j in zip_longest(changes[0::2], changes[1::2]):
            if (i, j) == (True, False) or (i, j) == (False, True):
                count_changes += 1

        if count_changes == 0:
            n_cluster = 1

        elif changes[0] and not changes[-1] or not changes[0] and changes[-1]:
            n_cluster = int(ceil(count_changes/2))

        elif changes[0] and changes[-1]:
            n_cluster = int(count_changes/2 + 1)

        else:
        # elif not changes[0] and not changes[-1]:
            n_cluster = int(count_changes/2)

        time_column = list(cutted_block.get_column("T"))
        return ckmeans(time_column, n_cluster)


# TODO (dmt): Refactor Ckmeans algorithm!
# Resource: https://journal.r-project.org/archive/2011-2/RJournal_2011-2_Wang+Song.pdf
def ssq(j, i, sum_x, sum_x_sq):
    if j > 0:
        muji = (sum_x[i] - sum_x[j-1]) / (i - j + 1)
        sji = sum_x_sq[i] - sum_x_sq[j-1] - (i - j + 1) * muji ** 2
    else:
        sji = sum_x_sq[i] - sum_x[i] ** 2 / (i+1)

    return 0 if sji < 0 else sji


def fill_row_k(imin, imax, k, S, J, sum_x, sum_x_sq, N):
    if imin > imax: return

    i = (imin+imax) // 2
    S[k][i] = S[k-1][i-1]
    J[k][i] = i

    jlow = k

    if imin > k:
        jlow = int(max(jlow, J[k][imin-1]))
    jlow = int(max(jlow, J[k-1][i]))

    jhigh = i-1
    if imax < N-1:
        jhigh = int(min(jhigh, J[k][imax+1]))

    for j in range(jhigh, jlow-1, -1):
        sji = ssq(j, i, sum_x, sum_x_sq)

        if sji + S[k-1][jlow-1] >= S[k][i]: break

        # Examine the lower bound of the cluster border
        # compute s(jlow, i)
        sjlowi = ssq(jlow, i, sum_x, sum_x_sq)

        SSQ_jlow = sjlowi + S[k-1][jlow-1]

        if SSQ_jlow < S[k][i]:
            S[k][i] = SSQ_jlow
            J[k][i] = jlow

        jlow += 1

        SSQ_j = sji + S[k-1][j-1]
        if SSQ_j < S[k][i]:
            S[k][i] = SSQ_j
            J[k][i] = j

    fill_row_k(imin, i-1, k, S, J, sum_x, sum_x_sq, N)
    fill_row_k(i+1, imax, k, S, J, sum_x, sum_x_sq, N)


def fill_dp_matrix(data, S, J, K, N):
    import numpy as np
    sum_x = np.zeros(N, dtype=np.float_)
    sum_x_sq = np.zeros(N, dtype=np.float_)

    # median. used to shift the values of x to improve numerical stability
    shift = data[N//2]

    for i in range(N):
        if i == 0:
            sum_x[0] = data[0] - shift
            sum_x_sq[0] = (data[0] - shift) ** 2
        else:
            sum_x[i] = sum_x[i-1] + data[i] - shift
            sum_x_sq[i] = sum_x_sq[i-1] + (data[i] - shift) ** 2

        S[0][i] = ssq(0, i, sum_x, sum_x_sq)
        J[0][i] = 0

    for k in range(1, K):
        if k < K-1:
            imin = max(1, k)
        else:
            imin = N-1

        fill_row_k(imin, N-1, k, S, J, sum_x, sum_x_sq, N)


def ckmeans(data, n_clusters):
    import numpy as np
    if n_clusters <= 0:
        raise ValueError("Cannot classify into 0 or less clusters")
    if n_clusters > len(data):
        raise ValueError("Cannot generate more classes than there are data values")

    # if there's only one value, return it; there's no sensible way to split
    # it. This means that len(ckmeans([data], 2)) may not == 2. Is that OK?
    unique = len(set(data))
    if unique == 1:
        return [data]

    data.sort()
    n = len(data)

    S = np.zeros((n_clusters, n), dtype=np.float_)

    J = np.zeros((n_clusters, n), dtype=np.uint64)

    fill_dp_matrix(data, S, J, n_clusters, n)

    clusters = []
    cluster_right = n-1

    for cluster in range(n_clusters-1, -1, -1):
        cluster_left = int(J[cluster][cluster_right])
        clusters.append(data[cluster_left:cluster_right+1])

        if cluster > 0:
            cluster_right = cluster_left - 1

    return list(reversed(clusters))
