import pickle
from collections import deque
from collections import defaultdict

from cml.domain.reconstruction import PragmaticMachineLearningModel


TS_QUEUE = deque()


def notify_inverted_index(func):
    def wrapper(self, *args, **kwargs):
        for obs in self.observer:
            getattr(obs, func.__name__)(*args, **kwargs)
        return func(self, *args, **kwargs)
    return wrapper


class KnowledgeDatabase:
    def __init__(self, highest_tier):
        self.database = [KnowledgeDomain(i) for i in range(highest_tier)]
        self.observer = []
        self.n_models = 0
        self.sources = []

    def __str__(self):
        return str(self.database)

    def __contains__(self, item: PragmaticMachineLearningModel):
        if not isinstance(item, PragmaticMachineLearningModel):
            raise TypeError()
        try:
            self.get(item.uid)
            return True
        except KeyError:
            return False

    @notify_inverted_index
    def insert(self, model: PragmaticMachineLearningModel):
        if model not in self:
            self.database[model.tier].insert(model)
            self.n_models += 1

    @notify_inverted_index
    def remove(self, model: PragmaticMachineLearningModel):
        self.database[model.tier].remove(model)
        self.n_models -= 1

    @notify_inverted_index
    def replace(self,
                replaced: PragmaticMachineLearningModel,
                replacer: PragmaticMachineLearningModel):
        self.database[replaced.tier].replace(replaced, replacer)

    @notify_inverted_index
    def extend(self): pass

    def get(self, uid: str):
        _, tier, _ = uid.split(".")
        return self.database[int(tier)].get(uid)

    def dump(self, path: str):
        file = open(path, "wb")
        pickle.dump(self, file)
        file.close()

    def model_counter(self):
        def counts(tier):
            return self.database[tier].counter()
        return counts

    def remove_dependent_models(self,
                                relative_model: PragmaticMachineLearningModel):
        to_remove = []
        for domain in self.database:
            for model in domain.knowledge.values():
                # TODO (dmt): Fix origin datatype.
                if isinstance(model.origin, str):
                    if model.origin == relative_model.uid:
                        to_remove.append(model)
                elif isinstance(model.origin, list):
                    if relative_model.uid in model.origin:
                        to_remove.append(model)
                elif isinstance(model.origin, tuple):
                    if relative_model.uid in model.origin:
                        to_remove.append(model)

                else: continue

        return to_remove
        # for model in to_remove:
        #    self.remove(model)

    def last_index_source(self):
        if self.sources:
            last_data_source = self.sources[-1]
            return last_data_source.source.indices()[-1]
        else:
            return 0

    def insert_source(self, source):
        self.sources.append(source)


class KnowledgeDomain:
    def __init__(self, tier: int):
        self.tier = tier
        self.knowledge = {}
        self.biggest_id = 0

    def __str__(self):
        return "\n".join(["{:~^100}".format(self.tier), "{}".format(
            self.knowledge)])

    def get(self, uid: str):
        return self.knowledge[uid]

    def insert(self, model: PragmaticMachineLearningModel):
        self.knowledge[model] = model
        #self.update_biggest_id(model.identifier)

    def remove(self, model: PragmaticMachineLearningModel):
        del self.knowledge[model]

    def replace(self,
                replaced: PragmaticMachineLearningModel,
                replacer: PragmaticMachineLearningModel):
        del self.knowledge[replaced]
        self.knowledge[replacer] = replacer
        #self.update_biggest_id(replacer.identifier)

    def counter(self):
        self.biggest_id += 1
        return self.biggest_id

class RelativeFinder:
    def __init__(self, tiers):
        self.indices = {}
        for tier in range(tiers):
            self.indices[tier] = (
                defaultdict(set),
                defaultdict(set),
                defaultdict(set))

    def find(self, pair, model):
        index_t, index_z, index_sigma = self.get_index_tier(model)
        if pair == ("T", "Z"):
            set_t = index_t[(model.min_timestamp, model.max_timestamp)]
            set_z = index_z[model.purpose]
            relatives = set_t.intersection(set_z)
        elif pair == ("T", "Sigma"):
            set_t = index_t[(model.min_timestamp, model.max_timestamp)]
            set_s = index_sigma[model.subject]
            relatives = set_t.intersection(set_s)
        elif pair == ("Sigma", "Z"):
            set_s = index_sigma[model.subject]
            set_z = index_z[model.purpose]
            relatives = set_s.intersection(set_z)
        elif pair == ("complete", ):
            set_t = set()
            for min_time, max_time in index_t.keys():
                if (model.min_timestamp >= min_time and
                    model.max_timestamp <= max_time) or \
                   (model.min_timestamp <= min_time and
                    model.max_timestamp >= max_time):
                    set_t.update(index_t[(min_time, max_time)])
            set_s = index_sigma[model.subject]
            set_z = index_z[model.purpose]
            relatives = set_t.intersection(set_s).intersection(set_z)
        else: raise ValueError

        if model in relatives:
            relatives.remove(model)

        return relatives

    def remove(self, model: PragmaticMachineLearningModel):
        t_list, s_list, z_list = self.get_index_lists(model,
                                                      time=True,
                                                      subject=True,
                                                      purpose=True)
        s_list.remove(model)
        z_list.remove(model)
        t_list.remove(model)

    def replace(self, replaced, replacer):
        t_list, s_list, z_list = self.get_index_lists(replaced,
                                                      time=True,
                                                      subject=True,
                                                      purpose=True)
        t_list.remove(replaced)
        s_list.remove(replaced)
        z_list.remove(replaced)
        t, z, sigma = self.get_index_tier(replacer)
        t[(replacer.min_timestamp, replacer.max_timestamp)].add(replacer)
        z[replacer.purpose].add(replacer)
        sigma[replacer.subject].add(replacer)

    def insert(self, model):
        t, z, sigma = self.get_index_tier(model)
        t[(model.min_timestamp, model.max_timestamp)].add(model)
        z[model.purpose].add(model)
        sigma[model.subject].add(model)

    def get_index_lists(self,
                        model: PragmaticMachineLearningModel,
                        time: bool,
                        subject: bool,
                        purpose: bool):
        t, z, sigma = self.get_index_tier(model)
        t_list = s_list = z_list = None
        if time:
            t_list = t[(model.min_timestamp, model.max_timestamp)]
        if subject:
            s_list = sigma[model.subject]
        if purpose:
            z_list = z[model.purpose]
        return t_list, s_list, z_list

    def get_index_tier(self, model):
        index_t = self.indices[model.meta.knowledge_tier][0]
        index_z = self.indices[model.meta.knowledge_tier][1]
        index_sigma = self.indices[model.meta.knowledge_tier][2]
        return index_t, index_z, index_sigma

    def extend(self): pass
