import json
from os.path import join
from enum import Enum
from functools import partial


class ConstructionInfo:
    class State(Enum):
        DISCARDED = 1
        CONSTRUCTED = 2
        PASSED = 3

    def __init__(self):
        self.n_cluster = None
        self.subject = None
        self.state = None


class Logger:
    JOURNAL = {}
    CURRENT_TIER = 1
    CURRENT_N_BLOCK = 1
    CURRENT_CONSTRUCTION_SUBJECT = None

    def log_tier(self, tier):
        Logger.JOURNAL[tier] = {}

    def log_n_block(self, counter):
        if Logger.CURRENT_TIER == 1:
            Logger.JOURNAL[self.CURRENT_TIER][counter] = {}
            Logger.CURRENT_N_BLOCK = counter

    def log_block_generation(self, block):
        if Logger.CURRENT_TIER == 1:
            Logger.JOURNAL[self.CURRENT_TIER][Logger.CURRENT_N_BLOCK]["BLOCK"] = {
                "rows": str(block.n_rows()),
                "columns": str(block.n_columns()),
                "origin": str(block.origin),
            }

        else:
            Logger.JOURNAL[Logger.CURRENT_TIER]["LEARNBLOCK"] = {
                "rows": str(block.n_columns()),
                "columns": str(block.n_rows()),
                "origin": str(block.origin),
                "CONSTRUCTION": {}
            }

    def log_lb_identification(self, lb):
        if Logger.CURRENT_TIER == 1:
            block = Logger.JOURNAL[self.CURRENT_TIER][Logger.CURRENT_N_BLOCK]["BLOCK"]
            block["LEARNBLOCK"] = {
                "relationship": str(lb.relationship),
                "CONSTRUCTION": {}
            }

    def log_construction(self, log):
        if log.state == ConstructionInfo.State.DISCARDED:
            return

        elif Logger.CURRENT_TIER == 1:
            Logger.CURRENT_CONSTRUCTION_SUBJECT = str(log.subject)+str(log.n_cluster)
            block = Logger.JOURNAL[self.CURRENT_TIER][Logger.CURRENT_N_BLOCK]["BLOCK"]
            block["LEARNBLOCK"]["CONSTRUCTION"].update({
                Logger.CURRENT_CONSTRUCTION_SUBJECT: {"status": str(log.state)}
            })

        else:
            Logger.CURRENT_CONSTRUCTION_SUBJECT = str(log.subject) + str(
                log.n_cluster)
            Logger.JOURNAL[Logger.CURRENT_TIER]["LEARNBLOCK"][
                "CONSTRUCTION"].update({
                Logger.CURRENT_CONSTRUCTION_SUBJECT: {"status": str(log.state)}
                }
            )

    def log_feature_selection(self, reduced):
        if Logger.CURRENT_TIER == 1:
            block = Logger.JOURNAL[self.CURRENT_TIER][Logger.CURRENT_N_BLOCK]["BLOCK"]
            construction = block["LEARNBLOCK"]["CONSTRUCTION"][Logger.CURRENT_CONSTRUCTION_SUBJECT]
            construction["SELECTION"] = {
                "rows": reduced.n_rows(),
                "columns": reduced.n_columns()
            }

    def log_reconstruction(self, winner):
        if Logger.CURRENT_TIER == 1:
            block = Logger.JOURNAL[self.CURRENT_TIER][Logger.CURRENT_N_BLOCK]["BLOCK"]
            construction = block["LEARNBLOCK"]["CONSTRUCTION"][Logger.CURRENT_CONSTRUCTION_SUBJECT]
            construction["RECONSTRUCTION"] = {
                "subject": str(winner.subject),
                "accuracy": str(winner.accuracy),
                "reliability": str(winner.reliability),
                "DECONSTRUCTION": {}
            }

        else:
            const = Logger.JOURNAL[Logger.CURRENT_TIER]["LEARNBLOCK"]["CONSTRUCTION"]
            construction = const[Logger.CURRENT_CONSTRUCTION_SUBJECT]
            construction["RECONSTRUCTION"] = {
                "subject": str(winner.subject),
                "accuracy": str(winner.accuracy),
                "reliability": str(winner.reliability),
                "DECONSTRUCTION": {}
            }

    def log_deconstruction(self, log):
        if Logger.CURRENT_TIER == 1:
            block = Logger.JOURNAL[self.CURRENT_TIER][Logger.CURRENT_N_BLOCK]["BLOCK"]
            construction = block["LEARNBLOCK"]["CONSTRUCTION"][Logger.CURRENT_CONSTRUCTION_SUBJECT]
            construction["RECONSTRUCTION"]["DECONSTRUCTION"][str(log.relative)] = {}
            deconstruction = construction["RECONSTRUCTION"]["DECONSTRUCTION"][str(log.relative)]

        else:
            const = Logger.JOURNAL[Logger.CURRENT_TIER]["LEARNBLOCK"]["CONSTRUCTION"]
            recon = const[Logger.CURRENT_CONSTRUCTION_SUBJECT]["RECONSTRUCTION"]
            recon["DECONSTRUCTION"][str(log.relative)] = {}
            deconstruction = recon["DECONSTRUCTION"][str(log.relative)]

        if log.relationship == ("complete", ) or log.relationship == ("Sigma", "Z"):
            deconstruction[str(log.relationship)] = {
                    "pragmatic": str(log.pragmatic),
                    "relative": str(log.relative),
                    "inserted": str(log.inserted),
                    "deleted": str(log.deleted),
                    "RECONSTRUCTION": []}

            for recon_info in log.reconstructed_info:
                deconstruction[str(log.relationship)]["RECONSTRUCTION"].append(
                    {"uid": str(recon_info),
                     "subject": str(recon_info.subject),
                     "accuracy": str(recon_info.accuracy),
                     "reliability": str(recon_info.reliability)})

        elif log.relationship == ("Sigma", "T"):
            deconstruction[str(log.relationship)] = {
                "pragmatic": str(log.pragmatic),
                "relative": str(log.relative),
                "Learnblock": {
                    "rows": str(log.new_learnblock.n_rows()),
                    "columns": str(log.new_learnblock.n_columns()),
                    "min T": str(log.new_learnblock.min_timestamp()),
                    "max T": str(log.new_learnblock.max_timestamp()),
                    "origin": str(log.new_learnblock.origin),
                    "subject": str(log.new_learnblock.subject)}}
        else:
            deconstruction["failed"] = {
                "pragmatic": str(log.pragmatic),
                "inserted": str(log.pragmatic),
                "deleted": str(log.deleted)
            }


class TierBanner(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func
        self.logger = None

    def __get__(self, owner, func):
        if not self.logger:
            self.logger = owner.logger

        return partial(self.__call__, owner)

    def __call__(self, owner):
        old_tier = 1
        self.log(old_tier)
        self.log_tier(old_tier)
        for tier, lb in self.func(owner):
            yield tier, lb

    def log(self, tier):
        self.logger.protocol("{:#^100}".format("TIER {}".format(tier)))


class BlockLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func
        self.logger = None

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger

        return partial(self.__call__, owner)

    def __call__(self, owner):
        for j, block in enumerate(self.func(owner), start=1):
            self.log(j, block)
            self.log_n_block(j)
            self.log_block_generation(block)
            yield block

    def log(self, counter, block):
        self.logger.protocol("{:*^100}".format("Block Generation"))
        self.logger.protocol("{:<20}{}".format("#:", counter))
        self.logger.protocol("{:<20}{}".format("Rows:", str(block.n_rows())))
        self.logger.protocol("{:<20}{}".format("Columns:", str(block.n_columns())))
        self.logger.protocol("{:<20}{}".format("Origin:", str(block.origin)))


class HighBlockLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func
        self.logger = None
        self.__log_destination = None

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger

        if self.__log_destination != owner.log_destination:
            self.__log_destination = owner.log_destination
        return partial(self.__call__, owner)

    def __call__(self, owner):
        old_tier = 2
        self.log_tier(old_tier)
        self.log_banner(old_tier)
        Logger.CURRENT_TIER = old_tier

        for tier, block in self.func(owner):
            if tier > old_tier:
                self.log_tier(tier)
                self.log_banner(tier)
                Logger.CURRENT_TIER = tier
                old_tier = tier

            self.log_block_generation(block)
            self.log(block)
            yield tier, block

        self.dump_journal()

    def log_banner(self, tier):
        self.logger.protocol("{:#^100}".format("TIER {}".format(tier)))

    def log(self, block):
        self.logger.protocol("{:*^100}".format("Block Generation"))
        self.logger.protocol("{:<20}{}".format("Rows:", str(block.n_rows())))
        self.logger.protocol("{:<20}{}".format("Columns:", str(block.n_columns())))
        self.logger.protocol("{:<20}{}".format("Origin:", str(block.origin)))

    def dump_journal(self):
        file = join(self.__log_destination, "log.json")
        with open(file, "w") as fp:
            json.dump(super().JOURNAL, fp)


class LearnblockLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func
        self.logger = None

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, block):
        lb = self.func(owner, block)
        self.log(lb)
        self.log_lb_identification(lb)
        return lb

    def log(self, lb):
        self.logger.protocol("{:*^100}".format("Learnblock identification"))
        self.logger.protocol("{:<20}{}".format("Rows:", str(lb.n_rows())))
        self.logger.protocol("{:<20}{}".format("Columns:", str(lb.n_columns())))
        self.logger.protocol("{:<20}{}".format("Origin:", str(lb.origin)))
        self.logger.protocol("{:<20}{}".format("Relationship", str(lb.relationship)))


class SelectorLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func
        self.logger = None

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, learnblock):
        reduced = self.func(owner, learnblock)
        self.log(reduced)
        self.log_feature_selection(reduced)
        return reduced

    def log(self, lb):
        self.logger.protocol("{:*^100}".format("Feature selection"))
        self.logger.protocol("{:<20}{}".format("rows:", str(lb.n_rows())))
        self.logger.protocol("{:<20}{}".format("columns:", str(lb.n_columns())))


class ConstLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func
        self.logger = None

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, learnblock):
        for info, complete_learnblock in self.func(owner, learnblock):
            self.log_construction(info)

            self.log_begin_construction()
            if info.state == ConstructionInfo.State.PASSED and self.logger:
                self.logger.protocol("{: <20}{}".format("State:", "PASSED"))
                self.log_learnglock_info(learnblock, complete_learnblock)

            elif info.state == ConstructionInfo.State.DISCARDED and self.logger:
                self.logger.protocol("{: <20}{}".format("State:", "DISCARDED"))

            elif info.state == ConstructionInfo.State.CONSTRUCTED and self.logger:
                self.logger.protocol("{: <20}{}".format("State:", "CONSTRUCTED"))
                self.log_learnglock_info(learnblock, complete_learnblock)

            if complete_learnblock is None:
                continue
            else:
                yield info, complete_learnblock

    def log_begin_construction(self):
        if self.logger:
            self.logger.protocol("{:*^100}".format("Construction"))

    def log_learnglock_info(self, learnblock, complete_learnblock):
        self.logger.protocol(
            "\t".join(["",
                       "[{:^5}{:^5}]".format(learnblock.n_rows(),
                                             learnblock.n_columns()),
                       str(complete_learnblock.min_timestamp()),
                       str(complete_learnblock.max_timestamp()),
                       "{:<20}".format(str(complete_learnblock.origin)),
                       "{:<20}".format(str(complete_learnblock.relationship)),
                       "{:<20}".format(str(complete_learnblock.subject)),
                       ]))

    def log_constructed_info(self, logging_info):
        self.logger.protocol("State: {:<20}".format("CONSTRUCTED"))
        self.logger.protocol("{:<20}".format(logging_info.subject))
        self.logger.protocol("{:<20}".format(logging_info.n_cluster))


class ReconstLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func
        self.logger = None

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, tier, learnblock):
        self.log_begin_construction()
        winner = self.func(owner, tier, learnblock)
        self.log_winner(winner)
        self.log_reconstruction(winner)
        return winner

    def log_winner(self, winner):
        if self.logger:
            self.logger.protocol("{: <20}{}".format("Reconstructed:", str(winner)))
            self.logger.protocol("{: <20}{: <15}{}".format("", "Subject:", str(winner.subject)))
            self.logger.protocol("{: <20}{: <15}{}".format("", "Accuracy:", str(winner.accuracy)))
            self.logger.protocol("{: <20}{: <15}{}".format("", "Reliability:",str(winner.reliability)))

    def log_begin_construction(self):
        if self.logger:
            self.logger.protocol("{:*^100}".format("Reconstruction"))


class DeconstInfo:
    def __init__(self):
        self.status = False
        self.relationship = None
        self.relative = None
        self.pragmatic = None
        self.reconstructed_info = []
        self.deleted = []
        self.inserted = []
        self.new_learnblock = None


class DeconstLogger(Logger):
    def __init__(self, func):
        super().__init__()
        self.func = func
        self.logger = None

    def __get__(self, owner, func):
        if self.logger is None:
            self.logger = owner.logger
        return partial(self.__call__, owner)

    def __call__(self, owner, tier, pragmatic, reduced):
        self.logger.protocol("{:*^100}".format("Deconstruction"))
        for log_info in self.func(owner, tier, pragmatic, reduced):
            self.log(log_info)
            self.log_deconstruction(log_info)

    def log(self, info):
        if info.relationship:
            self.logger.protocol("{:+^100}".format(str(info.relationship)))
            self.logger.protocol("{: <20}{}".format("Pragmatic:", str(info.pragmatic)))
            self.logger.protocol("{: <20}{}".format("Relative:", str(info.relative)))

            for r_info in info.reconstructed_info:
                self.log_reconstruction(r_info)

        self.logger.protocol("{: <20}{}".format("Inserted:", str(info.inserted)))
        self.logger.protocol("{: <20}{}".format("Deleted:", str(info.deleted)))

    def log_reconstruction(self, info):
        self.logger.protocol("{: <20}{}".format("Reconstructed:", str(info)))
        self.logger.protocol("{: <20}{: <15} {}".format("", "Subject:", str(info.subject)))
        self.logger.protocol("{: <20}{: <15} {}".format("", "Accuracy:", str(info.accuracy)))
        self.logger.protocol("{: <20}{: <15} {}".format("", "Reliability:", str(info.reliability)))

    def t_sigma_log(self, info):
        self.logger.protocol("{:+^100}".format(str(info.logs.relationship)))
        self.logger.protocol("{: <20}{}".format("Pragmatic:", str(info.logs.pragmatic)))
        self.logger.protocol("{: <20}{}".format("Relative:", str(info.relative)))
        if info.status:
            self.logger.protocol("{: <20}[{:^5}{:^5}]".format("Dimension: ", str(info.new_learnblock.n_rows()), str(info.new_learnblock.n_columns())))
            self.logger.protocol("{: <20}{}".format("Min T: ", info.new_learnblock.min_timestamp()))
            self.logger.protocol("{: <20}{}".format("Max T: ", info.new_learnblock.max_timestamp()))
            self.logger.protocol("{: <20}{}".format("Origin: ", str(info.new_learnblock.origin)))
            self.logger.protocol("{: <20}{}".format("Subject:", str(info.new_learnblock.subject)))
