from typing import List, Tuple

from cml.usecases.usecase import Usecase
from cml.domain import (
    Preprocessor,
    LearnblockIdentifier,
    DataSource,
    FeatureSelector,
    Reconstructor,
    Constructor,
    Deconstructor)
from cml.domain.knowledge import (
    KnowledgeDatabase,
    RelativeFinder)


class PreprocessUsecase(Usecase):
    def process(self, request):
        # Preprocess the data
        data = request.source_adapter.read_csv(request.general_settings.input_file)
        preprocesser = Preprocessor(request.preprocessing_settings)
        cleaned_data = preprocesser.clean(data)

        # Initialize learnblock identifier
        density_estimator = request.density_estimator_interface()
        # relative_extrema_finder = request.relative_extrema_interface
        settings = request.block_processing_settings
        learnblock_identifier = LearnblockIdentifier(settings,
                                                     density_estimator)
        return DataSource(cleaned_data, learnblock_identifier, settings)


class FeatureSelectionUsecase(Usecase):
    def process(self, request):
        return FeatureSelector(request.filter_method,
                               request.embedded_method,
                               request.settings)


class ReconstructionUsecase(Usecase):
    def process(self, request):
        models = create_models(request.interface, request.algorithms)
        return Reconstructor(request.settings, models, request.mode)


class ConstructionUsecase(Usecase):
    def process(self, request):
        models = create_models(request.interface, request.algorithms)
        constructor = Constructor(request.mode, models, request.settings)
        constructor.construction_type = request.mode
        return constructor


class DeconstructionUsecase(Usecase):
    def process(self, request):
        # TODO (dmt): Read an existing database.
        database = self.knowledge_database(request)
        relative_finder = RelativeFinder(request.settings.highest_tier)
        database.observer.append(relative_finder)
        return Deconstructor(request.mode,
                             database,
                             relative_finder,
                             request.source_adapter,
                             request.settings)

    def knowledge_database(self, request):
        if request.mode == "conceptual" and not request.knowledge_database:
            database = KnowledgeDatabase(request.settings.highest_tier)

        elif request.mode == "procedural" and not request.knowledge_database:
            database = KnowledgeDatabase(request.settings.highest_tier)

        elif request.knowledge_database and \
            (isinstance(request.knowledge_database, KnowledgeDatabase)
             and request.mode == "conceptual") or \
                (isinstance(request.knowledge_database, KnowledgeDatabase)
             and request.mode == "procedural"):
            database = request.knowledge_database
        else:
            raise ValueError

        return database


def create_models(interface: type, algorithms: List[Tuple]):
    models = []
    for abbreviation, algorithm in algorithms:
        model = interface(algorithm, abbreviation)
        models.append(model)
    return models

