from os import mkdir
from os.path import expanduser, join
import multiprocessing
from datetime import datetime
from logging import getLogger, FileHandler, StreamHandler
from contextlib import suppress

from cml.shared.parameter import PROTOCOL_LEVEL
from cml.usecases.usecase import Usecase
from cml.domain.knowledge import TS_QUEUE
from cml.shared.errors import (
    ReconstructionFailure,
    SelectionFailure,
    ModeError)
from cml.shared.logger import (
    BlockLogger, HighBlockLogger,
    LearnblockLogger,
    TierBanner,
    SelectorLogger,
    ConstLogger,
    ReconstLogger,
    DeconstLogger)


class BlockGeneratorWrapper:
    def __init__(self, source, logger):
        self.source = source
        self.logger = logger

    def prepare(self, old_source_size, start_index, end_index):
        self.source.n_source = old_source_size
        self.source.reindex(start_index, end_index)
        self.source.remap_feature_names()

    @BlockLogger
    def blocks(self):
        for block in self.source.blocks:
            yield block

    def update_generator(self, lb):
        self.source._flip_source_halde_flags(lb.indices())

    def get_origin(self):
        return self.source.n_source


class LearnblockIdentifierWrapper:
    def __init__(self, identifier, logger):
        self.identifier = identifier
        self.logger = logger

    def prepare(self):
        pass

    @LearnblockLogger
    def identify(self, block):
        learnblock =  self.identifier.identify(block)
        return learnblock


class SourceWrapper:
    def __init__(self, source_wrapper, identifier_wrapper, logger):
        self.source_wrapper = source_wrapper
        self.identifier_wrapper = identifier_wrapper
        self.logger = logger
        self.log_destination = None

    def prepare(self, log_dir):
        self.log_destination = log_dir

    def learnblocks(self):
        for tier, lb in self.tier_one_learnblocks():
            yield tier, lb

        for tier, lb in self.tier_n_learnblocks():
            yield tier, lb

    @TierBanner
    def tier_one_learnblocks(self):
        first_tier = 1
        for block in self.source_wrapper.blocks():
            lb = self.identifier_wrapper.identify(block)
            if lb: lb.origin = ("source", self.source_wrapper.get_origin())
            self.source_wrapper.update_generator(lb)
            yield first_tier, lb

    @HighBlockLogger
    def tier_n_learnblocks(self):
        while TS_QUEUE:
            tier, lb = TS_QUEUE.pop()
            if tier and lb:
                yield tier, lb

    @property
    def source(self):
        return self.source_wrapper.source


class ConstructorWrapper:
    def __init__(self, con, logger):
        self.con = con
        self.logger = logger
        self.act = None

    def prepare(self, parallel=False):
        if parallel:
            self.act = self.con.construct_parallel
        else:
            self.act = self.con.construct_serial

    @ConstLogger
    def construct(self, learnblock):
        for log_info, labeled in self.act(learnblock):
            yield log_info, labeled


class FeatureSelectorWrapper:
    def __init__(self, selector, logger):
        self.selector = selector
        self.logger = logger

    def prepare(self):
        pass

    @SelectorLogger
    def select(self, labeled):
        return self.selector.select(labeled)


class ReconstructorWrapper:
    def __init__(self, recon, logger):
        self.recon = recon
        self.logger = logger
        self.act = None

    def prepare(self, model_counter, parallel=False):
        self.recon.inject_counter(model_counter)

        if parallel:
            self.act = self.recon.reconstruct_parallel
        else:
            self.act = self.recon.reconstruct_serial

    @ReconstLogger
    def reconstruct(self, tier, reduced):
        return self.act(tier, reduced)


class DeconstructorWrapper:
    def __init__(self, decon, logger):
        self.decon = decon
        self.logger = logger
        self.act = None

    def prepare(self, recon, parallel=False):
        self.decon.reconstructor = recon
        if parallel:
            self.act = self.decon.deconstruct_parallel
        else:
            self.act = self.decon.deconstruct_serial

    @DeconstLogger
    def deconstruct(self, tier, pragmatic, reduced):
        for log_info in self.act(tier, pragmatic, reduced):
            yield log_info

    def finish(self, source):
        self.decon.knowledge_database.insert_source(source)


class KnowledgeSearchUsecase(Usecase):
    FIRST_TIER = 1

    def __init__(self):
        super().__init__()
        self.logger = None

    def process(self, request):
        self.valid_constraints(request)
        wrappers = self.prepare_wrappers(request)
        self.search(wrappers)
        return self.finish_wrappers(wrappers)

    def valid_constraints(self, req):
        if not self.valid_modi(req):
            raise ModeError()

    def prepare_wrappers(self, req):
        logger, log_dir = self.prepare_logger(req)
        self.logger = logger
        model_counter = req.deconstructor.knowledge_database.model_counter()

        # Prepare data generator
        n_source = len(req.deconstructor.knowledge_database.sources)
        start_index = req.deconstructor.knowledge_database.last_index_source()
        end_index = start_index+len(req.deconstructor.source)
        generator = BlockGeneratorWrapper(req.deconstructor.source, logger)
        generator.prepare(n_source, start_index, end_index)

        # Prepare lb identifier
        identifier = LearnblockIdentifierWrapper(req.deconstructor.source.identifier, logger)

        # Prepare data source
        source = SourceWrapper(generator, identifier, logger)
        source.prepare(log_dir)

        # Prepare constructor
        con = ConstructorWrapper(req.constructor, logger)
        con.prepare(req.parallel)

        # Prepare selector
        selector = FeatureSelectorWrapper(req.feature_selector, logger)

        # Prepare reconstructor
        recon = ReconstructorWrapper(req.reconstructor, logger)
        recon.prepare(model_counter, req.parallel)

        # Prepare deconstructor
        decon = DeconstructorWrapper(req.deconstructor, logger)
        decon.prepare(req.reconstructor, req.parallel)

        return source, con, selector, recon, decon

    def prepare_logger(self, req):
        log_dir = create_log_dir()
        iteration_log_file = join(log_dir, "iteration.log")
        iteration_handler = FileHandler(iteration_log_file)
        iteration_handler.level = PROTOCOL_LEVEL
        iteration_logger = getLogger("iterationLogger")

        # Remove loggers from other sessions
        if iteration_logger.handlers: iteration_logger.handlers.clear()
        iteration_logger.addHandler(iteration_handler)

        if req.stdout:
            iteration_stdout_handler = StreamHandler()
            iteration_stdout_handler.level = PROTOCOL_LEVEL
            iteration_logger.addHandler(iteration_stdout_handler)
        return iteration_logger, log_dir

    def search(self, wrappers):
        for tier, lb in wrappers[0].learnblocks():
            for _, labeled in wrappers[1].construct(lb):
                with suppress(SelectionFailure, ReconstructionFailure):
                    reduced = wrappers[2].select(labeled)
                    pragmatic = wrappers[3].reconstruct(tier, reduced)
                    wrappers[4].deconstruct(tier, pragmatic, reduced)

    def finish_wrappers(self, wrappers):
        wrappers[4].finish(wrappers[0].source)
        for child in multiprocessing.active_children():
            child.kill()

        return wrappers[4].decon.knowledge_database

    def valid_modi(self, request):
        return request.constructor.mode == request.reconstructor.mode == \
               request.deconstructor.mode


def create_log_dir():
    home_folder_path = expanduser("~")
    cml_folder = ".cml"
    current_time = datetime.now().strftime("%d%m%Y%H%M%S")
    log_dir = join(home_folder_path, cml_folder, current_time)
    mkdir(log_dir)
    return log_dir

